@@ -17,7 +17,6 @@ import (
 	"context"
 	"encoding/json"
 	"io"
-	"io/ioutil"
 	"os"
 	"path/filepath"
 	"sort"
@@ -26,11 +25,16 @@ import (
 	"time"
 
 	"github.com/prometheus/common/model"
+	"github.com/stretchr/testify/require"
+	"go.uber.org/goleak"
 
 	"github.com/prometheus/prometheus/discovery/targetgroup"
-	"github.com/prometheus/prometheus/util/testutil"
 )
 
+func TestMain(m *testing.M) {
+	goleak.VerifyTestMain(m)
+}
+
 const defaultWait = time.Second
 
 type testRunner struct {
@@ -48,12 +52,9 @@ type testRunner struct {
 func newTestRunner(t *testing.T) *testRunner {
 	t.Helper()
 
-	tmpDir, err := ioutil.TempDir("", "prometheus-file-sd")
-	testutil.Ok(t, err)
-
 	return &testRunner{
 		T:       t,
-		dir:     tmpDir,
+		dir:     t.TempDir(),
 		ch:      make(chan []*targetgroup.Group),
 		done:    make(chan struct{}),
 		stopped: make(chan struct{}),
@@ -68,40 +69,40 @@ func (t *testRunner) copyFile(src string) string {
 }
 
 // copyFileTo atomically copies a file with a different name to the runner's directory.
-func (t *testRunner) copyFileTo(src string, name string) string {
+func (t *testRunner) copyFileTo(src, name string) string {
 	t.Helper()
 
-	newf, err := ioutil.TempFile(t.dir, "")
-	testutil.Ok(t, err)
+	newf, err := os.CreateTemp(t.dir, "")
+	require.NoError(t, err)
 
 	f, err := os.Open(src)
-	testutil.Ok(t, err)
+	require.NoError(t, err)
 
 	_, err = io.Copy(newf, f)
-	testutil.Ok(t, err)
-	testutil.Ok(t, f.Close())
-	testutil.Ok(t, newf.Close())
+	require.NoError(t, err)
+	require.NoError(t, f.Close())
+	require.NoError(t, newf.Close())
 
 	dst := filepath.Join(t.dir, name)
 	err = os.Rename(newf.Name(), dst)
-	testutil.Ok(t, err)
+	require.NoError(t, err)
 
 	return dst
 }
 
 // writeString writes atomically a string to a file.
-func (t *testRunner) writeString(file string, data string) {
+func (t *testRunner) writeString(file, data string) {
 	t.Helper()
 
-	newf, err := ioutil.TempFile(t.dir, "")
-	testutil.Ok(t, err)
+	newf, err := os.CreateTemp(t.dir, "")
+	require.NoError(t, err)
 
 	_, err = newf.WriteString(data)
-	testutil.Ok(t, err)
-	testutil.Ok(t, newf.Close())
+	require.NoError(t, err)
+	require.NoError(t, newf.Close())
 
 	err = os.Rename(newf.Name(), file)
-	testutil.Ok(t, err)
+	require.NoError(t, err)
 }
 
 // appendString appends a string to a file.
@@ -109,11 +110,11 @@ func (t *testRunner) appendString(file, data string) {
 	t.Helper()
 
 	f, err := os.OpenFile(file, os.O_WRONLY|os.O_APPEND, 0)
-	testutil.Ok(t, err)
+	require.NoError(t, err)
 	defer f.Close()
 
 	_, err = f.WriteString(data)
-	testutil.Ok(t, err)
+	require.NoError(t, err)
 }
 
 // run starts the file SD and the loop receiving target groups updates.
@@ -225,7 +226,7 @@ func (t *testRunner) requireTargetGroups(expected, got []*targetgroup.Group) {
 		panic(err)
 	}
 
-	testutil.Equals(t, string(b1), string(b2))
+	require.Equal(t, string(b1), string(b2))
 }
 
 // validTg() maps to fixtures/valid.{json,yml}.
@@ -307,6 +308,7 @@ func TestInitialUpdate(t *testing.T) {
 		"fixtures/valid.yml",
 		"fixtures/valid.json",
 	} {
+		tc := tc
 		t.Run(tc, func(t *testing.T) {
 			t.Parallel()
 
@@ -463,7 +465,7 @@ func TestRemoveFile(t *testing.T) {
 
 	// Verify that we receive the update about the target groups being removed.
 	ref := runner.lastReceive()
-	testutil.Ok(t, os.Remove(sdFile))
+	require.NoError(t, os.Remove(sdFile))
 	runner.requireUpdate(
 		ref,
 		[]*targetgroup.Group{
@@ -472,6 +474,7 @@ func TestRemoveFile(t *testing.T) {
 			},
 			{
 				Source: fileSource(sdFile, 1),
-			}},
+			},
+		},
 	)
 }
