@@ -18,10 +18,13 @@ import (
 	"encoding/binary"
 	"runtime"
 	"sort"
-	"strings"
 	"sync"
 
-	"github.com/prometheus/prometheus/pkg/labels"
+	"github.com/pkg/errors"
+	"golang.org/x/exp/slices"
+
+	"github.com/prometheus/prometheus/model/labels"
+	"github.com/prometheus/prometheus/storage"
 )
 
 var allPostingsKey = labels.Label{}
@@ -31,33 +34,67 @@ func AllPostingsKey() (name, value string) {
 	return allPostingsKey.Name, allPostingsKey.Value
 }
 
+// ensureOrderBatchSize is the max number of postings passed to a worker in a single batch in MemPostings.EnsureOrder().
+const ensureOrderBatchSize = 1024
+
+// ensureOrderBatchPool is a pool used to recycle batches passed to workers in MemPostings.EnsureOrder().
+var ensureOrderBatchPool = sync.Pool{
+	New: func() interface{} {
+		x := make([][]storage.SeriesRef, 0, ensureOrderBatchSize)
+		return &x // Return pointer type as preferred by Pool.
+	},
+}
+
 // MemPostings holds postings list for series ID per label pair. They may be written
 // to out of order.
-// ensureOrder() must be called once before any reads are done. This allows for quick
+// EnsureOrder() must be called once before any reads are done. This allows for quick
 // unordered batch fills on startup.
 type MemPostings struct {
 	mtx     sync.RWMutex
-	m       map[string]map[string][]uint64
+	m       map[string]map[string][]storage.SeriesRef
 	ordered bool
 }
 
 // NewMemPostings returns a memPostings that's ready for reads and writes.
 func NewMemPostings() *MemPostings {
 	return &MemPostings{
-		m:       make(map[string]map[string][]uint64, 512),
+		m:       make(map[string]map[string][]storage.SeriesRef, 512),
 		ordered: true,
 	}
 }
 
 // NewUnorderedMemPostings returns a memPostings that is not safe to be read from
-// until ensureOrder was called once.
+// until EnsureOrder() was called once.
 func NewUnorderedMemPostings() *MemPostings {
 	return &MemPostings{
-		m:       make(map[string]map[string][]uint64, 512),
+		m:       make(map[string]map[string][]storage.SeriesRef, 512),
 		ordered: false,
 	}
 }
 
+// Symbols returns an iterator over all unique name and value strings, in order.
+func (p *MemPostings) Symbols() StringIter {
+	p.mtx.RLock()
+
+	// Add all the strings to a map to de-duplicate.
+	symbols := make(map[string]struct{}, 512)
+	for n, e := range p.m {
+		symbols[n] = struct{}{}
+		for v := range e {
+			symbols[v] = struct{}{}
+		}
+	}
+	p.mtx.RUnlock()
+
+	res := make([]string, 0, len(symbols))
+	for k := range symbols {
+		res = append(res, k)
+	}
+
+	slices.Sort(res)
+	return NewStringListIter(res)
+}
+
 // SortedKeys returns a list of sorted label keys of the postings.
 func (p *MemPostings) SortedKeys() []labels.Label {
 	p.mtx.RLock()
@@ -71,51 +108,83 @@ func (p *MemPostings) SortedKeys() []labels.Label {
 	p.mtx.RUnlock()
 
 	sort.Slice(keys, func(i, j int) bool {
-		if d := strings.Compare(keys[i].Name, keys[j].Name); d != 0 {
-			return d < 0
+		if keys[i].Name != keys[j].Name {
+			return keys[i].Name < keys[j].Name
 		}
 		return keys[i].Value < keys[j].Value
 	})
 	return keys
 }
 
+// LabelNames returns all the unique label names.
+func (p *MemPostings) LabelNames() []string {
+	p.mtx.RLock()
+	defer p.mtx.RUnlock()
+	n := len(p.m)
+	if n == 0 {
+		return nil
+	}
+
+	names := make([]string, 0, n-1)
+	for name := range p.m {
+		if name != allPostingsKey.Name {
+			names = append(names, name)
+		}
+	}
+	return names
+}
+
+// LabelValues returns label values for the given name.
+func (p *MemPostings) LabelValues(name string) []string {
+	p.mtx.RLock()
+	defer p.mtx.RUnlock()
+
+	values := make([]string, 0, len(p.m[name]))
+	for v := range p.m[name] {
+		values = append(values, v)
+	}
+	return values
+}
+
 // PostingsStats contains cardinality based statistics for postings.
 type PostingsStats struct {
 	CardinalityMetricsStats []Stat
 	CardinalityLabelStats   []Stat
 	LabelValueStats         []Stat
 	LabelValuePairsStats    []Stat
+	NumLabelPairs           int
 }
 
 // Stats calculates the cardinality statistics from postings.
-func (p *MemPostings) Stats(label string) *PostingsStats {
-	const maxNumOfRecords = 10
+func (p *MemPostings) Stats(label string, limit int) *PostingsStats {
 	var size uint64
-
 	p.mtx.RLock()
 
 	metrics := &maxHeap{}
 	labels := &maxHeap{}
 	labelValueLength := &maxHeap{}
 	labelValuePairs := &maxHeap{}
+	numLabelPairs := 0
 
-	metrics.init(maxNumOfRecords)
-	labels.init(maxNumOfRecords)
-	labelValueLength.init(maxNumOfRecords)
-	labelValuePairs.init(maxNumOfRecords)
+	metrics.init(limit)
+	labels.init(limit)
+	labelValueLength.init(limit)
+	labelValuePairs.init(limit)
 
 	for n, e := range p.m {
 		if n == "" {
 			continue
 		}
 		labels.push(Stat{Name: n, Count: uint64(len(e))})
+		numLabelPairs += len(e)
 		size = 0
 		for name, values := range e {
 			if n == label {
 				metrics.push(Stat{Name: name, Count: uint64(len(values))})
 			}
-			labelValuePairs.push(Stat{Name: n + "=" + name, Count: uint64(len(values))})
-			size += uint64(len(name))
+			seriesCnt := uint64(len(values))
+			labelValuePairs.push(Stat{Name: n + "=" + name, Count: seriesCnt})
+			size += uint64(len(name)) * seriesCnt
 		}
 		labelValueLength.push(Stat{Name: n, Count: size})
 	}
@@ -127,12 +196,13 @@ func (p *MemPostings) Stats(label string) *PostingsStats {
 		CardinalityLabelStats:   labels.get(),
 		LabelValueStats:         labelValueLength.get(),
 		LabelValuePairsStats:    labelValuePairs.get(),
+		NumLabelPairs:           numLabelPairs,
 	}
 }
 
 // Get returns a postings list for the given label pair.
 func (p *MemPostings) Get(name, value string) Postings {
-	var lp []uint64
+	var lp []storage.SeriesRef
 	p.mtx.RLock()
 	l := p.m[name]
 	if l != nil {
@@ -153,7 +223,10 @@ func (p *MemPostings) All() Postings {
 
 // EnsureOrder ensures that all postings lists are sorted. After it returns all further
 // calls to add and addFor will insert new IDs in a sorted manner.
-func (p *MemPostings) EnsureOrder() {
+// Parameter numberOfConcurrentProcesses is used to specify the maximal number of
+// CPU cores used for this operation. If it is <= 0, GOMAXPROCS is used.
+// GOMAXPROCS was the default before introducing this parameter.
+func (p *MemPostings) EnsureOrder(numberOfConcurrentProcesses int) {
 	p.mtx.Lock()
 	defer p.mtx.Unlock()
 
@@ -161,26 +234,46 @@ func (p *MemPostings) EnsureOrder() {
 		return
 	}
 
-	n := runtime.GOMAXPROCS(0)
-	workc := make(chan []uint64)
+	concurrency := numberOfConcurrentProcesses
+	if concurrency <= 0 {
+		concurrency = runtime.GOMAXPROCS(0)
+	}
+	workc := make(chan *[][]storage.SeriesRef)
 
 	var wg sync.WaitGroup
-	wg.Add(n)
+	wg.Add(concurrency)
 
-	for i := 0; i < n; i++ {
+	for i := 0; i < concurrency; i++ {
 		go func() {
-			for l := range workc {
-				sort.Slice(l, func(a, b int) bool { return l[a] < l[b] })
+			for job := range workc {
+				for _, l := range *job {
+					slices.Sort(l)
+				}
+
+				*job = (*job)[:0]
+				ensureOrderBatchPool.Put(job)
 			}
 			wg.Done()
 		}()
 	}
 
+	nextJob := ensureOrderBatchPool.Get().(*[][]storage.SeriesRef)
 	for _, e := range p.m {
 		for _, l := range e {
-			workc <- l
+			*nextJob = append(*nextJob, l)
+
+			if len(*nextJob) >= ensureOrderBatchSize {
+				workc <- nextJob
+				nextJob = ensureOrderBatchPool.Get().(*[][]storage.SeriesRef)
+			}
 		}
 	}
+
+	// If the last job was partially filled, we need to push it to workers too.
+	if len(*nextJob) > 0 {
+		workc <- nextJob
+	}
+
 	close(workc)
 	wg.Wait()
 
@@ -188,7 +281,7 @@ func (p *MemPostings) EnsureOrder() {
 }
 
 // Delete removes all ids in the given map from the postings lists.
-func (p *MemPostings) Delete(deleted map[uint64]struct{}) {
+func (p *MemPostings) Delete(deleted map[storage.SeriesRef]struct{}) {
 	var keys, vals []string
 
 	// Collect all keys relevant for deletion once. New keys added afterwards
@@ -224,7 +317,7 @@ func (p *MemPostings) Delete(deleted map[uint64]struct{}) {
 				p.mtx.Unlock()
 				continue
 			}
-			repl := make([]uint64, 0, len(p.m[n][l]))
+			repl := make([]storage.SeriesRef, 0, len(p.m[n][l]))
 
 			for _, id := range p.m[n][l] {
 				if _, ok := deleted[id]; !ok {
@@ -262,21 +355,21 @@ func (p *MemPostings) Iter(f func(labels.Label, Postings) error) error {
 }
 
 // Add a label set to the postings index.
-func (p *MemPostings) Add(id uint64, lset labels.Labels) {
+func (p *MemPostings) Add(id storage.SeriesRef, lset labels.Labels) {
 	p.mtx.Lock()
 
-	for _, l := range lset {
+	lset.Range(func(l labels.Label) {
 		p.addFor(id, l)
-	}
+	})
 	p.addFor(id, allPostingsKey)
 
 	p.mtx.Unlock()
 }
 
-func (p *MemPostings) addFor(id uint64, l labels.Label) {
+func (p *MemPostings) addFor(id storage.SeriesRef, l labels.Label) {
 	nm, ok := p.m[l.Name]
 	if !ok {
-		nm = map[string][]uint64{}
+		nm = map[string][]storage.SeriesRef{}
 		p.m[l.Name] = nm
 	}
 	list := append(nm[l.Value], id)
@@ -298,7 +391,7 @@ func (p *MemPostings) addFor(id uint64, l labels.Label) {
 }
 
 // ExpandPostings returns the postings expanded as a slice.
-func ExpandPostings(p Postings) (res []uint64, err error) {
+func ExpandPostings(p Postings) (res []storage.SeriesRef, err error) {
 	for p.Next() {
 		res = append(res, p.At())
 	}
@@ -312,10 +405,10 @@ type Postings interface {
 
 	// Seek advances the iterator to value v or greater and returns
 	// true if a value was found.
-	Seek(v uint64) bool
+	Seek(v storage.SeriesRef) bool
 
 	// At returns the value at the current iterator position.
-	At() uint64
+	At() storage.SeriesRef
 
 	// Err returns the last error of the iterator.
 	Err() error
@@ -326,20 +419,27 @@ type errPostings struct {
 	err error
 }
 
-func (e errPostings) Next() bool       { return false }
-func (e errPostings) Seek(uint64) bool { return false }
-func (e errPostings) At() uint64       { return 0 }
-func (e errPostings) Err() error       { return e.err }
+func (e errPostings) Next() bool                  { return false }
+func (e errPostings) Seek(storage.SeriesRef) bool { return false }
+func (e errPostings) At() storage.SeriesRef       { return 0 }
+func (e errPostings) Err() error                  { return e.err }
 
 var emptyPostings = errPostings{}
 
 // EmptyPostings returns a postings list that's always empty.
-// NOTE: Returning EmptyPostings sentinel when index.Postings struct has no postings is recommended.
+// NOTE: Returning EmptyPostings sentinel when Postings struct has no postings is recommended.
 // It triggers optimized flow in other functions like Intersect, Without etc.
 func EmptyPostings() Postings {
 	return emptyPostings
 }
 
+// IsEmptyPostingsType returns true if the postings are an empty postings list.
+// When this function returns false, it doesn't mean that the postings isn't empty
+// (it could be an empty intersection of two non-empty postings, for example).
+func IsEmptyPostingsType(p Postings) bool {
+	return p == emptyPostings
+}
+
 // ErrPostings returns new postings that immediately error.
 func ErrPostings(err error) Postings {
 	return errPostings{err}
@@ -365,14 +465,14 @@ func Intersect(its ...Postings) Postings {
 
 type intersectPostings struct {
 	arr []Postings
-	cur uint64
+	cur storage.SeriesRef
 }
 
 func newIntersectPostings(its ...Postings) *intersectPostings {
 	return &intersectPostings{arr: its}
 }
 
-func (it *intersectPostings) At() uint64 {
+func (it *intersectPostings) At() storage.SeriesRef {
 	return it.cur
 }
 
@@ -404,7 +504,7 @@ func (it *intersectPostings) Next() bool {
 	return it.doNext()
 }
 
-func (it *intersectPostings) Seek(id uint64) bool {
+func (it *intersectPostings) Seek(id storage.SeriesRef) bool {
 	it.cur = id
 	return it.doNext()
 }
@@ -455,7 +555,7 @@ func (h *postingsHeap) Pop() interface{} {
 type mergedPostings struct {
 	h           postingsHeap
 	initialized bool
-	cur         uint64
+	cur         storage.SeriesRef
 	err         error
 }
 
@@ -464,12 +564,11 @@ func newMergedPostings(p []Postings) (m *mergedPostings, nonEmpty bool) {
 
 	for _, it := range p {
 		// NOTE: mergedPostings struct requires the user to issue an initial Next.
-		if it.Next() {
+		switch {
+		case it.Next():
 			ph = append(ph, it)
-		} else {
-			if it.Err() != nil {
-				return &mergedPostings{err: it.Err()}, true
-			}
+		case it.Err() != nil:
+			return &mergedPostings{err: it.Err()}, true
 		}
 	}
 
@@ -515,7 +614,7 @@ func (it *mergedPostings) Next() bool {
 	}
 }
 
-func (it *mergedPostings) Seek(id uint64) bool {
+func (it *mergedPostings) Seek(id storage.SeriesRef) bool {
 	if it.h.Len() == 0 || it.err != nil {
 		return false
 	}
@@ -545,7 +644,7 @@ func (it *mergedPostings) Seek(id uint64) bool {
 	return true
 }
 
-func (it mergedPostings) At() uint64 {
+func (it mergedPostings) At() storage.SeriesRef {
 	return it.cur
 }
 
@@ -569,7 +668,7 @@ func Without(full, drop Postings) Postings {
 type removedPostings struct {
 	full, remove Postings
 
-	cur uint64
+	cur storage.SeriesRef
 
 	initialized bool
 	fok, rok    bool
@@ -582,7 +681,7 @@ func newRemovedPostings(full, remove Postings) *removedPostings {
 	}
 }
 
-func (rp *removedPostings) At() uint64 {
+func (rp *removedPostings) At() storage.SeriesRef {
 	return rp.cur
 }
 
@@ -602,24 +701,23 @@ func (rp *removedPostings) Next() bool {
 			rp.fok = rp.full.Next()
 			return true
 		}
-
-		fcur, rcur := rp.full.At(), rp.remove.At()
-		if fcur < rcur {
+		switch fcur, rcur := rp.full.At(), rp.remove.At(); {
+		case fcur < rcur:
 			rp.cur = fcur
 			rp.fok = rp.full.Next()
 
 			return true
-		} else if rcur < fcur {
+		case rcur < fcur:
 			// Forward the remove postings to the right position.
 			rp.rok = rp.remove.Seek(fcur)
-		} else {
+		default:
 			// Skip the current posting.
 			rp.fok = rp.full.Next()
 		}
 	}
 }
 
-func (rp *removedPostings) Seek(id uint64) bool {
+func (rp *removedPostings) Seek(id storage.SeriesRef) bool {
 	if rp.cur >= id {
 		return true
 	}
@@ -641,19 +739,19 @@ func (rp *removedPostings) Err() error {
 
 // ListPostings implements the Postings interface over a plain list.
 type ListPostings struct {
-	list []uint64
-	cur  uint64
+	list []storage.SeriesRef
+	cur  storage.SeriesRef
 }
 
-func NewListPostings(list []uint64) Postings {
+func NewListPostings(list []storage.SeriesRef) Postings {
 	return newListPostings(list...)
 }
 
-func newListPostings(list ...uint64) *ListPostings {
+func newListPostings(list ...storage.SeriesRef) *ListPostings {
 	return &ListPostings{list: list}
 }
 
-func (it *ListPostings) At() uint64 {
+func (it *ListPostings) At() storage.SeriesRef {
 	return it.cur
 }
 
@@ -667,7 +765,7 @@ func (it *ListPostings) Next() bool {
 	return false
 }
 
-func (it *ListPostings) Seek(x uint64) bool {
+func (it *ListPostings) Seek(x storage.SeriesRef) bool {
 	// If the current value satisfies, then return.
 	if it.cur >= x {
 		return true
@@ -704,8 +802,8 @@ func newBigEndianPostings(list []byte) *bigEndianPostings {
 	return &bigEndianPostings{list: list}
 }
 
-func (it *bigEndianPostings) At() uint64 {
-	return uint64(it.cur)
+func (it *bigEndianPostings) At() storage.SeriesRef {
+	return storage.SeriesRef(it.cur)
 }
 
 func (it *bigEndianPostings) Next() bool {
@@ -717,8 +815,8 @@ func (it *bigEndianPostings) Next() bool {
 	return false
 }
 
-func (it *bigEndianPostings) Seek(x uint64) bool {
-	if uint64(it.cur) >= x {
+func (it *bigEndianPostings) Seek(x storage.SeriesRef) bool {
+	if storage.SeriesRef(it.cur) >= x {
 		return true
 	}
 
@@ -740,3 +838,122 @@ func (it *bigEndianPostings) Seek(x uint64) bool {
 func (it *bigEndianPostings) Err() error {
 	return nil
 }
+
+// FindIntersectingPostings checks the intersection of p and candidates[i] for each i in candidates,
+// if intersection is non empty, then i is added to the indexes returned.
+// Returned indexes are not sorted.
+func FindIntersectingPostings(p Postings, candidates []Postings) (indexes []int, err error) {
+	h := make(postingsWithIndexHeap, 0, len(candidates))
+	for idx, it := range candidates {
+		switch {
+		case it.Next():
+			h = append(h, postingsWithIndex{index: idx, p: it})
+		case it.Err() != nil:
+			return nil, it.Err()
+		}
+	}
+	if h.empty() {
+		return nil, nil
+	}
+	heap.Init(&h)
+
+	for !h.empty() {
+		if !p.Seek(h.at()) {
+			return indexes, p.Err()
+		}
+		if p.At() == h.at() {
+			indexes = append(indexes, h.popIndex())
+		} else if err := h.next(); err != nil {
+			return nil, err
+		}
+	}
+
+	return indexes, nil
+}
+
+// postingsWithIndex is used as postingsWithIndexHeap elements by FindIntersectingPostings,
+// keeping track of the original index of each postings while they move inside the heap.
+type postingsWithIndex struct {
+	index int
+	p     Postings
+	// popped means that these postings shouldn't be considered anymore.
+	// See popIndex() comment to understand why we need this.
+	popped bool
+}
+
+// postingsWithIndexHeap implements heap.Interface,
+// with root always pointing to the postings with minimum Postings.At() value.
+// It also implements a special way of removing elements that marks them as popped and moves them to the bottom of the
+// heap instead of actually removing them, see popIndex() for more details.
+type postingsWithIndexHeap []postingsWithIndex
+
+// empty checks whether the heap is empty, which is true if it has no elements, of if the smallest element is popped.
+func (h *postingsWithIndexHeap) empty() bool {
+	return len(*h) == 0 || (*h)[0].popped
+}
+
+// popIndex pops the smallest heap element and returns its index.
+// In our implementation we don't actually do heap.Pop(), instead we mark the element as `popped` and fix its position, which
+// should be after all the non-popped elements according to our sorting strategy.
+// By skipping the `heap.Pop()` call we avoid an extra allocation in this heap's Pop() implementation which returns an interface{}.
+func (h *postingsWithIndexHeap) popIndex() int {
+	index := (*h)[0].index
+	(*h)[0].popped = true
+	heap.Fix(h, 0)
+	return index
+}
+
+// at provides the storage.SeriesRef where root Postings is pointing at this moment.
+func (h postingsWithIndexHeap) at() storage.SeriesRef { return h[0].p.At() }
+
+// next performs the Postings.Next() operation on the root of the heap, performing the related operation on the heap
+// and conveniently returning the result of calling Postings.Err() if the result of calling Next() was false.
+// If Next() succeeds, heap is fixed to move the root to its new position, according to its Postings.At() value.
+// If Next() returns fails and there's no error reported by Postings.Err(), then root is marked as removed and heap is fixed.
+func (h *postingsWithIndexHeap) next() error {
+	pi := (*h)[0]
+	next := pi.p.Next()
+	if next {
+		heap.Fix(h, 0)
+		return nil
+	}
+
+	if err := pi.p.Err(); err != nil {
+		return errors.Wrapf(err, "postings %d", pi.index)
+	}
+	h.popIndex()
+	return nil
+}
+
+// Len implements heap.Interface.
+// Notice that Len() > 0 does not imply that heap is not empty as elements are not removed from this heap.
+// Use empty() to check whether heap is empty or not.
+func (h postingsWithIndexHeap) Len() int { return len(h) }
+
+// Less implements heap.Interface, it puts all the popped elements at the bottom,
+// and then sorts by Postings.At() property of each node.
+func (h postingsWithIndexHeap) Less(i, j int) bool {
+	if h[i].popped != h[j].popped {
+		return h[j].popped
+	}
+	return h[i].p.At() < h[j].p.At()
+}
+
+// Swap implements heap.Interface.
+func (h *postingsWithIndexHeap) Swap(i, j int) { (*h)[i], (*h)[j] = (*h)[j], (*h)[i] }
+
+// Push implements heap.Interface.
+func (h *postingsWithIndexHeap) Push(x interface{}) {
+	*h = append(*h, x.(postingsWithIndex))
+}
+
+// Pop implements heap.Interface and pops the last element, which is NOT the min element,
+// so this doesn't return the same heap.Pop()
+// Although this method is implemented for correctness, we don't expect it to be used, see popIndex() method for details.
+func (h *postingsWithIndexHeap) Pop() interface{} {
+	old := *h
+	n := len(old)
+	x := old[n-1]
+	*h = old[0 : n-1]
+	return x
+}
