@@ -14,41 +14,47 @@
 package index
 
 import (
+	"container/heap"
+	"context"
 	"encoding/binary"
 	"fmt"
 	"math/rand"
 	"sort"
+	"strconv"
 	"testing"
 
-	"github.com/prometheus/prometheus/pkg/labels"
-	"github.com/prometheus/prometheus/util/testutil"
+	"github.com/pkg/errors"
+	"github.com/stretchr/testify/require"
+
+	"github.com/prometheus/prometheus/model/labels"
+	"github.com/prometheus/prometheus/storage"
 )
 
 func TestMemPostings_addFor(t *testing.T) {
 	p := NewMemPostings()
-	p.m[allPostingsKey.Name] = map[string][]uint64{}
-	p.m[allPostingsKey.Name][allPostingsKey.Value] = []uint64{1, 2, 3, 4, 6, 7, 8}
+	p.m[allPostingsKey.Name] = map[string][]storage.SeriesRef{}
+	p.m[allPostingsKey.Name][allPostingsKey.Value] = []storage.SeriesRef{1, 2, 3, 4, 6, 7, 8}
 
 	p.addFor(5, allPostingsKey)
 
-	testutil.Equals(t, []uint64{1, 2, 3, 4, 5, 6, 7, 8}, p.m[allPostingsKey.Name][allPostingsKey.Value])
+	require.Equal(t, []storage.SeriesRef{1, 2, 3, 4, 5, 6, 7, 8}, p.m[allPostingsKey.Name][allPostingsKey.Value])
 }
 
 func TestMemPostings_ensureOrder(t *testing.T) {
 	p := NewUnorderedMemPostings()
-	p.m["a"] = map[string][]uint64{}
+	p.m["a"] = map[string][]storage.SeriesRef{}
 
 	for i := 0; i < 100; i++ {
-		l := make([]uint64, 100)
+		l := make([]storage.SeriesRef, 100)
 		for j := range l {
-			l[j] = rand.Uint64()
+			l[j] = storage.SeriesRef(rand.Uint64())
 		}
 		v := fmt.Sprintf("%d", i)
 
 		p.m["a"][v] = l
 	}
 
-	p.EnsureOrder()
+	p.EnsureOrder(0)
 
 	for _, e := range p.m {
 		for _, l := range e {
@@ -62,11 +68,64 @@ func TestMemPostings_ensureOrder(t *testing.T) {
 	}
 }
 
+func BenchmarkMemPostings_ensureOrder(b *testing.B) {
+	tests := map[string]struct {
+		numLabels         int
+		numValuesPerLabel int
+		numRefsPerValue   int
+	}{
+		"many values per label": {
+			numLabels:         100,
+			numValuesPerLabel: 10000,
+			numRefsPerValue:   100,
+		},
+		"few values per label": {
+			numLabels:         1000000,
+			numValuesPerLabel: 1,
+			numRefsPerValue:   100,
+		},
+		"few refs per label value": {
+			numLabels:         1000,
+			numValuesPerLabel: 1000,
+			numRefsPerValue:   10,
+		},
+	}
+
+	for testName, testData := range tests {
+		b.Run(testName, func(b *testing.B) {
+			p := NewUnorderedMemPostings()
+
+			// Generate postings.
+			for l := 0; l < testData.numLabels; l++ {
+				labelName := strconv.Itoa(l)
+				p.m[labelName] = map[string][]storage.SeriesRef{}
+
+				for v := 0; v < testData.numValuesPerLabel; v++ {
+					refs := make([]storage.SeriesRef, testData.numRefsPerValue)
+					for j := range refs {
+						refs[j] = storage.SeriesRef(rand.Uint64())
+					}
+
+					labelValue := strconv.Itoa(v)
+					p.m[labelName][labelValue] = refs
+				}
+			}
+
+			b.ResetTimer()
+
+			for n := 0; n < b.N; n++ {
+				p.EnsureOrder(0)
+				p.ordered = false
+			}
+		})
+	}
+}
+
 func TestIntersect(t *testing.T) {
 	a := newListPostings(1, 2, 3)
 	b := newListPostings(2, 3, 4)
 
-	var cases = []struct {
+	cases := []struct {
 		in []Postings
 
 		res Postings
@@ -160,12 +219,12 @@ func TestIntersect(t *testing.T) {
 			}
 
 			expected, err := ExpandPostings(c.res)
-			testutil.Ok(t, err)
+			require.NoError(t, err)
 
 			i := Intersect(c.in...)
 
 			if c.res == EmptyPostings() {
-				testutil.Equals(t, EmptyPostings(), i)
+				require.Equal(t, EmptyPostings(), i)
 				return
 			}
 
@@ -174,37 +233,37 @@ func TestIntersect(t *testing.T) {
 			}
 
 			res, err := ExpandPostings(i)
-			testutil.Ok(t, err)
-			testutil.Equals(t, expected, res)
+			require.NoError(t, err)
+			require.Equal(t, expected, res)
 		})
 	}
 }
 
 func TestMultiIntersect(t *testing.T) {
-	var cases = []struct {
-		p   [][]uint64
-		res []uint64
+	cases := []struct {
+		p   [][]storage.SeriesRef
+		res []storage.SeriesRef
 	}{
 		{
-			p: [][]uint64{
+			p: [][]storage.SeriesRef{
 				{1, 2, 3, 4, 5, 6, 1000, 1001},
 				{2, 4, 5, 6, 7, 8, 999, 1001},
 				{1, 2, 5, 6, 7, 8, 1001, 1200},
 			},
-			res: []uint64{2, 5, 6, 1001},
+			res: []storage.SeriesRef{2, 5, 6, 1001},
 		},
-		// One of the reproduceable cases for:
+		// One of the reproducible cases for:
 		// https://github.com/prometheus/prometheus/issues/2616
 		// The initialisation of intersectPostings was moving the iterator forward
 		// prematurely making us miss some postings.
 		{
-			p: [][]uint64{
+			p: [][]storage.SeriesRef{
 				{1, 2},
 				{1, 2},
 				{1, 2},
 				{2},
 			},
-			res: []uint64{2},
+			res: []storage.SeriesRef{2},
 		},
 	}
 
@@ -216,70 +275,75 @@ func TestMultiIntersect(t *testing.T) {
 
 		res, err := ExpandPostings(Intersect(ps...))
 
-		testutil.Ok(t, err)
-		testutil.Equals(t, c.res, res)
+		require.NoError(t, err)
+		require.Equal(t, c.res, res)
 	}
 }
 
+func consumePostings(p Postings) error {
+	for p.Next() {
+		p.At()
+	}
+	return p.Err()
+}
+
 func BenchmarkIntersect(t *testing.B) {
 	t.Run("LongPostings1", func(bench *testing.B) {
-		var a, b, c, d []uint64
+		var a, b, c, d []storage.SeriesRef
 
 		for i := 0; i < 10000000; i += 2 {
-			a = append(a, uint64(i))
+			a = append(a, storage.SeriesRef(i))
 		}
 		for i := 5000000; i < 5000100; i += 4 {
-			b = append(b, uint64(i))
+			b = append(b, storage.SeriesRef(i))
 		}
 		for i := 5090000; i < 5090600; i += 4 {
-			b = append(b, uint64(i))
+			b = append(b, storage.SeriesRef(i))
 		}
 		for i := 4990000; i < 5100000; i++ {
-			c = append(c, uint64(i))
+			c = append(c, storage.SeriesRef(i))
 		}
 		for i := 4000000; i < 6000000; i++ {
-			d = append(d, uint64(i))
+			d = append(d, storage.SeriesRef(i))
 		}
 
-		i1 := newListPostings(a...)
-		i2 := newListPostings(b...)
-		i3 := newListPostings(c...)
-		i4 := newListPostings(d...)
-
 		bench.ResetTimer()
 		bench.ReportAllocs()
 		for i := 0; i < bench.N; i++ {
-			if _, err := ExpandPostings(Intersect(i1, i2, i3, i4)); err != nil {
+			i1 := newListPostings(a...)
+			i2 := newListPostings(b...)
+			i3 := newListPostings(c...)
+			i4 := newListPostings(d...)
+			if err := consumePostings(Intersect(i1, i2, i3, i4)); err != nil {
 				bench.Fatal(err)
 			}
 		}
 	})
 
 	t.Run("LongPostings2", func(bench *testing.B) {
-		var a, b, c, d []uint64
+		var a, b, c, d []storage.SeriesRef
 
 		for i := 0; i < 12500000; i++ {
-			a = append(a, uint64(i))
+			a = append(a, storage.SeriesRef(i))
 		}
 		for i := 7500000; i < 12500000; i++ {
-			b = append(b, uint64(i))
+			b = append(b, storage.SeriesRef(i))
 		}
 		for i := 9000000; i < 20000000; i++ {
-			c = append(c, uint64(i))
+			c = append(c, storage.SeriesRef(i))
 		}
 		for i := 10000000; i < 12000000; i++ {
-			d = append(d, uint64(i))
+			d = append(d, storage.SeriesRef(i))
 		}
 
-		i1 := newListPostings(a...)
-		i2 := newListPostings(b...)
-		i3 := newListPostings(c...)
-		i4 := newListPostings(d...)
-
 		bench.ResetTimer()
 		bench.ReportAllocs()
 		for i := 0; i < bench.N; i++ {
-			if _, err := ExpandPostings(Intersect(i1, i2, i3, i4)); err != nil {
+			i1 := newListPostings(a...)
+			i2 := newListPostings(b...)
+			i3 := newListPostings(c...)
+			i4 := newListPostings(d...)
+			if err := consumePostings(Intersect(i1, i2, i3, i4)); err != nil {
 				bench.Fatal(err)
 			}
 		}
@@ -287,21 +351,29 @@ func BenchmarkIntersect(t *testing.B) {
 
 	// Many matchers(k >> n).
 	t.Run("ManyPostings", func(bench *testing.B) {
-		var its []Postings
+		var lps []*ListPostings
+		var refs [][]storage.SeriesRef
 
-		// 100000 matchers(k=100000).
+		// Create 100000 matchers(k=100000), making sure all memory allocation is done before starting the loop.
 		for i := 0; i < 100000; i++ {
-			var temp []uint64
-			for j := 1; j < 100; j++ {
-				temp = append(temp, uint64(j))
+			var temp []storage.SeriesRef
+			for j := storage.SeriesRef(1); j < 100; j++ {
+				temp = append(temp, j)
 			}
-			its = append(its, newListPostings(temp...))
+			lps = append(lps, newListPostings(temp...))
+			refs = append(refs, temp)
 		}
 
+		its := make([]Postings, len(refs))
 		bench.ResetTimer()
 		bench.ReportAllocs()
 		for i := 0; i < bench.N; i++ {
-			if _, err := ExpandPostings(Intersect(its...)); err != nil {
+			// Reset the ListPostings to their original values each time round the loop.
+			for j := range refs {
+				lps[j].list = refs[j]
+				its[j] = lps[j]
+			}
+			if err := consumePostings(Intersect(its...)); err != nil {
 				bench.Fatal(err)
 			}
 		}
@@ -314,12 +386,12 @@ func TestMultiMerge(t *testing.T) {
 	i3 := newListPostings(1, 2, 5, 6, 7, 8, 1001, 1200)
 
 	res, err := ExpandPostings(Merge(i1, i2, i3))
-	testutil.Ok(t, err)
-	testutil.Equals(t, []uint64{1, 2, 3, 4, 5, 6, 7, 8, 999, 1000, 1001, 1200}, res)
+	require.NoError(t, err)
+	require.Equal(t, []storage.SeriesRef{1, 2, 3, 4, 5, 6, 7, 8, 999, 1000, 1001, 1200}, res)
 }
 
 func TestMergedPostings(t *testing.T) {
-	var cases = []struct {
+	cases := []struct {
 		in []Postings
 
 		res Postings
@@ -402,12 +474,12 @@ func TestMergedPostings(t *testing.T) {
 			}
 
 			expected, err := ExpandPostings(c.res)
-			testutil.Ok(t, err)
+			require.NoError(t, err)
 
 			m := Merge(c.in...)
 
 			if c.res == EmptyPostings() {
-				testutil.Equals(t, EmptyPostings(), m)
+				require.Equal(t, EmptyPostings(), m)
 				return
 			}
 
@@ -416,51 +488,51 @@ func TestMergedPostings(t *testing.T) {
 			}
 
 			res, err := ExpandPostings(m)
-			testutil.Ok(t, err)
-			testutil.Equals(t, expected, res)
+			require.NoError(t, err)
+			require.Equal(t, expected, res)
 		})
 	}
 }
 
 func TestMergedPostingsSeek(t *testing.T) {
-	var cases = []struct {
-		a, b []uint64
+	cases := []struct {
+		a, b []storage.SeriesRef
 
-		seek    uint64
+		seek    storage.SeriesRef
 		success bool
-		res     []uint64
+		res     []storage.SeriesRef
 	}{
 		{
-			a: []uint64{2, 3, 4, 5},
-			b: []uint64{6, 7, 8, 9, 10},
+			a: []storage.SeriesRef{2, 3, 4, 5},
+			b: []storage.SeriesRef{6, 7, 8, 9, 10},
 
 			seek:    1,
 			success: true,
-			res:     []uint64{2, 3, 4, 5, 6, 7, 8, 9, 10},
+			res:     []storage.SeriesRef{2, 3, 4, 5, 6, 7, 8, 9, 10},
 		},
 		{
-			a: []uint64{1, 2, 3, 4, 5},
-			b: []uint64{6, 7, 8, 9, 10},
+			a: []storage.SeriesRef{1, 2, 3, 4, 5},
+			b: []storage.SeriesRef{6, 7, 8, 9, 10},
 
 			seek:    2,
 			success: true,
-			res:     []uint64{2, 3, 4, 5, 6, 7, 8, 9, 10},
+			res:     []storage.SeriesRef{2, 3, 4, 5, 6, 7, 8, 9, 10},
 		},
 		{
-			a: []uint64{1, 2, 3, 4, 5},
-			b: []uint64{4, 5, 6, 7, 8},
+			a: []storage.SeriesRef{1, 2, 3, 4, 5},
+			b: []storage.SeriesRef{4, 5, 6, 7, 8},
 
 			seek:    9,
 			success: false,
 			res:     nil,
 		},
 		{
-			a: []uint64{1, 2, 3, 4, 9, 10},
-			b: []uint64{1, 4, 5, 6, 7, 8, 10, 11},
+			a: []storage.SeriesRef{1, 2, 3, 4, 9, 10},
+			b: []storage.SeriesRef{1, 4, 5, 6, 7, 8, 10, 11},
 
 			seek:    10,
 			success: true,
-			res:     []uint64{10, 11},
+			res:     []storage.SeriesRef{10, 11},
 		},
 	}
 
@@ -470,59 +542,59 @@ func TestMergedPostingsSeek(t *testing.T) {
 
 		p := Merge(a, b)
 
-		testutil.Equals(t, c.success, p.Seek(c.seek))
+		require.Equal(t, c.success, p.Seek(c.seek))
 
 		// After Seek(), At() should be called.
 		if c.success {
 			start := p.At()
 			lst, err := ExpandPostings(p)
-			testutil.Ok(t, err)
+			require.NoError(t, err)
 
-			lst = append([]uint64{start}, lst...)
-			testutil.Equals(t, c.res, lst)
+			lst = append([]storage.SeriesRef{start}, lst...)
+			require.Equal(t, c.res, lst)
 		}
 	}
 }
 
 func TestRemovedPostings(t *testing.T) {
-	var cases = []struct {
-		a, b []uint64
-		res  []uint64
+	cases := []struct {
+		a, b []storage.SeriesRef
+		res  []storage.SeriesRef
 	}{
 		{
 			a:   nil,
 			b:   nil,
-			res: []uint64(nil),
+			res: []storage.SeriesRef(nil),
 		},
 		{
-			a:   []uint64{1, 2, 3, 4},
+			a:   []storage.SeriesRef{1, 2, 3, 4},
 			b:   nil,
-			res: []uint64{1, 2, 3, 4},
+			res: []storage.SeriesRef{1, 2, 3, 4},
 		},
 		{
 			a:   nil,
-			b:   []uint64{1, 2, 3, 4},
-			res: []uint64(nil),
+			b:   []storage.SeriesRef{1, 2, 3, 4},
+			res: []storage.SeriesRef(nil),
 		},
 		{
-			a:   []uint64{1, 2, 3, 4, 5},
-			b:   []uint64{6, 7, 8, 9, 10},
-			res: []uint64{1, 2, 3, 4, 5},
+			a:   []storage.SeriesRef{1, 2, 3, 4, 5},
+			b:   []storage.SeriesRef{6, 7, 8, 9, 10},
+			res: []storage.SeriesRef{1, 2, 3, 4, 5},
 		},
 		{
-			a:   []uint64{1, 2, 3, 4, 5},
-			b:   []uint64{4, 5, 6, 7, 8},
-			res: []uint64{1, 2, 3},
+			a:   []storage.SeriesRef{1, 2, 3, 4, 5},
+			b:   []storage.SeriesRef{4, 5, 6, 7, 8},
+			res: []storage.SeriesRef{1, 2, 3},
 		},
 		{
-			a:   []uint64{1, 2, 3, 4, 9, 10},
-			b:   []uint64{1, 4, 5, 6, 7, 8, 10, 11},
-			res: []uint64{2, 3, 9},
+			a:   []storage.SeriesRef{1, 2, 3, 4, 9, 10},
+			b:   []storage.SeriesRef{1, 4, 5, 6, 7, 8, 10, 11},
+			res: []storage.SeriesRef{2, 3, 9},
 		},
 		{
-			a:   []uint64{1, 2, 3, 4, 9, 10},
-			b:   []uint64{1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11},
-			res: []uint64(nil),
+			a:   []storage.SeriesRef{1, 2, 3, 4, 9, 10},
+			b:   []storage.SeriesRef{1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11},
+			res: []storage.SeriesRef(nil),
 		},
 	}
 
@@ -531,17 +603,16 @@ func TestRemovedPostings(t *testing.T) {
 		b := newListPostings(c.b...)
 
 		res, err := ExpandPostings(newRemovedPostings(a, b))
-		testutil.Ok(t, err)
-		testutil.Equals(t, c.res, res)
+		require.NoError(t, err)
+		require.Equal(t, c.res, res)
 	}
-
 }
 
 func TestRemovedNextStackoverflow(t *testing.T) {
-	var full []uint64
-	var remove []uint64
+	var full []storage.SeriesRef
+	var remove []storage.SeriesRef
 
-	var i uint64
+	var i storage.SeriesRef
 	for i = 0; i < 1e7; i++ {
 		full = append(full, i)
 		remove = append(remove, i)
@@ -555,73 +626,73 @@ func TestRemovedNextStackoverflow(t *testing.T) {
 		gotElem = true
 	}
 
-	testutil.Ok(t, rp.Err())
-	testutil.Assert(t, !gotElem, "")
+	require.NoError(t, rp.Err())
+	require.False(t, gotElem)
 }
 
 func TestRemovedPostingsSeek(t *testing.T) {
-	var cases = []struct {
-		a, b []uint64
+	cases := []struct {
+		a, b []storage.SeriesRef
 
-		seek    uint64
+		seek    storage.SeriesRef
 		success bool
-		res     []uint64
+		res     []storage.SeriesRef
 	}{
 		{
-			a: []uint64{2, 3, 4, 5},
-			b: []uint64{6, 7, 8, 9, 10},
+			a: []storage.SeriesRef{2, 3, 4, 5},
+			b: []storage.SeriesRef{6, 7, 8, 9, 10},
 
 			seek:    1,
 			success: true,
-			res:     []uint64{2, 3, 4, 5},
+			res:     []storage.SeriesRef{2, 3, 4, 5},
 		},
 		{
-			a: []uint64{1, 2, 3, 4, 5},
-			b: []uint64{6, 7, 8, 9, 10},
+			a: []storage.SeriesRef{1, 2, 3, 4, 5},
+			b: []storage.SeriesRef{6, 7, 8, 9, 10},
 
 			seek:    2,
 			success: true,
-			res:     []uint64{2, 3, 4, 5},
+			res:     []storage.SeriesRef{2, 3, 4, 5},
 		},
 		{
-			a: []uint64{1, 2, 3, 4, 5},
-			b: []uint64{4, 5, 6, 7, 8},
+			a: []storage.SeriesRef{1, 2, 3, 4, 5},
+			b: []storage.SeriesRef{4, 5, 6, 7, 8},
 
 			seek:    9,
 			success: false,
 			res:     nil,
 		},
 		{
-			a: []uint64{1, 2, 3, 4, 9, 10},
-			b: []uint64{1, 4, 5, 6, 7, 8, 10, 11},
+			a: []storage.SeriesRef{1, 2, 3, 4, 9, 10},
+			b: []storage.SeriesRef{1, 4, 5, 6, 7, 8, 10, 11},
 
 			seek:    10,
 			success: false,
 			res:     nil,
 		},
 		{
-			a: []uint64{1, 2, 3, 4, 9, 10},
-			b: []uint64{1, 4, 5, 6, 7, 8, 11},
+			a: []storage.SeriesRef{1, 2, 3, 4, 9, 10},
+			b: []storage.SeriesRef{1, 4, 5, 6, 7, 8, 11},
 
 			seek:    4,
 			success: true,
-			res:     []uint64{9, 10},
+			res:     []storage.SeriesRef{9, 10},
 		},
 		{
-			a: []uint64{1, 2, 3, 4, 9, 10},
-			b: []uint64{1, 4, 5, 6, 7, 8, 11},
+			a: []storage.SeriesRef{1, 2, 3, 4, 9, 10},
+			b: []storage.SeriesRef{1, 4, 5, 6, 7, 8, 11},
 
 			seek:    5,
 			success: true,
-			res:     []uint64{9, 10},
+			res:     []storage.SeriesRef{9, 10},
 		},
 		{
-			a: []uint64{1, 2, 3, 4, 9, 10},
-			b: []uint64{1, 4, 5, 6, 7, 8, 11},
+			a: []storage.SeriesRef{1, 2, 3, 4, 9, 10},
+			b: []storage.SeriesRef{1, 4, 5, 6, 7, 8, 11},
 
 			seek:    10,
 			success: true,
-			res:     []uint64{10},
+			res:     []storage.SeriesRef{10},
 		},
 	}
 
@@ -631,16 +702,16 @@ func TestRemovedPostingsSeek(t *testing.T) {
 
 		p := newRemovedPostings(a, b)
 
-		testutil.Equals(t, c.success, p.Seek(c.seek))
+		require.Equal(t, c.success, p.Seek(c.seek))
 
 		// After Seek(), At() should be called.
 		if c.success {
 			start := p.At()
 			lst, err := ExpandPostings(p)
-			testutil.Ok(t, err)
+			require.NoError(t, err)
 
-			lst = append([]uint64{start}, lst...)
-			testutil.Equals(t, c.res, lst)
+			lst = append([]storage.SeriesRef{start}, lst...)
+			require.Equal(t, c.res, lst)
 		}
 	}
 }
@@ -663,12 +734,12 @@ func TestBigEndian(t *testing.T) {
 	t.Run("Iteration", func(t *testing.T) {
 		bep := newBigEndianPostings(beLst)
 		for i := 0; i < num; i++ {
-			testutil.Assert(t, bep.Next() == true, "")
-			testutil.Equals(t, uint64(ls[i]), bep.At())
+			require.True(t, bep.Next())
+			require.Equal(t, storage.SeriesRef(ls[i]), bep.At())
 		}
 
-		testutil.Assert(t, bep.Next() == false, "")
-		testutil.Assert(t, bep.Err() == nil, "")
+		require.False(t, bep.Next())
+		require.NoError(t, bep.Err())
 	})
 
 	t.Run("Seek", func(t *testing.T) {
@@ -712,9 +783,9 @@ func TestBigEndian(t *testing.T) {
 		bep := newBigEndianPostings(beLst)
 
 		for _, v := range table {
-			testutil.Equals(t, v.found, bep.Seek(uint64(v.seek)))
-			testutil.Equals(t, uint64(v.val), bep.At())
-			testutil.Assert(t, bep.Err() == nil, "")
+			require.Equal(t, v.found, bep.Seek(storage.SeriesRef(v.seek)))
+			require.Equal(t, storage.SeriesRef(v.val), bep.At())
+			require.NoError(t, bep.Err())
 		}
 	})
 }
@@ -732,12 +803,12 @@ func TestIntersectWithMerge(t *testing.T) {
 	p := Intersect(a, b)
 	res, err := ExpandPostings(p)
 
-	testutil.Ok(t, err)
-	testutil.Equals(t, []uint64{30}, res)
+	require.NoError(t, err)
+	require.Equal(t, []storage.SeriesRef{30}, res)
 }
 
 func TestWithoutPostings(t *testing.T) {
-	var cases = []struct {
+	cases := []struct {
 		base Postings
 		drop Postings
 
@@ -794,12 +865,12 @@ func TestWithoutPostings(t *testing.T) {
 			}
 
 			expected, err := ExpandPostings(c.res)
-			testutil.Ok(t, err)
+			require.NoError(t, err)
 
 			w := Without(c.base, c.drop)
 
 			if c.res == EmptyPostings() {
-				testutil.Equals(t, EmptyPostings(), w)
+				require.Equal(t, EmptyPostings(), w)
 				return
 			}
 
@@ -808,8 +879,8 @@ func TestWithoutPostings(t *testing.T) {
 			}
 
 			res, err := ExpandPostings(w)
-			testutil.Ok(t, err)
-			testutil.Equals(t, expected, res)
+			require.NoError(t, err)
+			require.Equal(t, expected, res)
 		})
 	}
 }
@@ -817,12 +888,14 @@ func TestWithoutPostings(t *testing.T) {
 func BenchmarkPostings_Stats(b *testing.B) {
 	p := NewMemPostings()
 
+	var seriesID storage.SeriesRef
+
 	createPostingsLabelValues := func(name, valuePrefix string, count int) {
 		for n := 1; n < count; n++ {
 			value := fmt.Sprintf("%s-%d", valuePrefix, n)
-			p.Add(uint64(n), labels.FromStrings(name, value))
+			p.Add(seriesID, labels.FromStrings(name, value))
+			seriesID++
 		}
-
 	}
 	createPostingsLabelValues("__name__", "metrics_name_can_be_very_big_and_bad", 1e3)
 	for i := 0; i < 20; i++ {
@@ -839,9 +912,37 @@ func BenchmarkPostings_Stats(b *testing.B) {
 	}
 	b.ResetTimer()
 	for n := 0; n < b.N; n++ {
-		p.Stats("__name__")
+		p.Stats("__name__", 10)
 	}
+}
 
+func TestMemPostingsStats(t *testing.T) {
+	// create a new MemPostings
+	p := NewMemPostings()
+
+	// add some postings to the MemPostings
+	p.Add(1, labels.FromStrings("label", "value1"))
+	p.Add(1, labels.FromStrings("label", "value2"))
+	p.Add(1, labels.FromStrings("label", "value3"))
+	p.Add(2, labels.FromStrings("label", "value1"))
+
+	// call the Stats method to calculate the cardinality statistics
+	stats := p.Stats("label", 10)
+
+	// assert that the expected statistics were calculated
+	require.Equal(t, uint64(2), stats.CardinalityMetricsStats[0].Count)
+	require.Equal(t, "value1", stats.CardinalityMetricsStats[0].Name)
+
+	require.Equal(t, uint64(3), stats.CardinalityLabelStats[0].Count)
+	require.Equal(t, "label", stats.CardinalityLabelStats[0].Name)
+
+	require.Equal(t, uint64(24), stats.LabelValueStats[0].Count)
+	require.Equal(t, "label", stats.LabelValueStats[0].Name)
+
+	require.Equal(t, uint64(2), stats.LabelValuePairsStats[0].Count)
+	require.Equal(t, "label=value1", stats.LabelValuePairsStats[0].Name)
+
+	require.Equal(t, 3, stats.NumLabelPairs)
 }
 
 func TestMemPostings_Delete(t *testing.T) {
@@ -851,25 +952,164 @@ func TestMemPostings_Delete(t *testing.T) {
 	p.Add(3, labels.FromStrings("lbl2", "a"))
 
 	before := p.Get(allPostingsKey.Name, allPostingsKey.Value)
-	p.Delete(map[uint64]struct{}{
-		2: struct{}{},
+	p.Delete(map[storage.SeriesRef]struct{}{
+		2: {},
 	})
 	after := p.Get(allPostingsKey.Name, allPostingsKey.Value)
 
 	// Make sure postings gotten before the delete have the old data when
 	// iterated over.
 	expanded, err := ExpandPostings(before)
-	testutil.Ok(t, err)
-	testutil.Equals(t, []uint64{1, 2, 3}, expanded)
+	require.NoError(t, err)
+	require.Equal(t, []storage.SeriesRef{1, 2, 3}, expanded)
 
 	// Make sure postings gotten after the delete have the new data when
 	// iterated over.
 	expanded, err = ExpandPostings(after)
-	testutil.Ok(t, err)
-	testutil.Equals(t, []uint64{1, 3}, expanded)
+	require.NoError(t, err)
+	require.Equal(t, []storage.SeriesRef{1, 3}, expanded)
 
 	deleted := p.Get("lbl1", "b")
 	expanded, err = ExpandPostings(deleted)
-	testutil.Ok(t, err)
-	testutil.Assert(t, 0 == len(expanded), "expected empty postings, got %v", expanded)
+	require.NoError(t, err)
+	require.Equal(t, 0, len(expanded), "expected empty postings, got %v", expanded)
+}
+
+func TestFindIntersectingPostings(t *testing.T) {
+	t.Run("multiple intersections", func(t *testing.T) {
+		p := NewListPostings([]storage.SeriesRef{10, 15, 20, 25, 30, 35, 40, 45, 50})
+		candidates := []Postings{
+			0: NewListPostings([]storage.SeriesRef{7, 13, 14, 27}), // Does not intersect.
+			1: NewListPostings([]storage.SeriesRef{10, 20}),        // Does intersect.
+			2: NewListPostings([]storage.SeriesRef{29, 30, 31}),    // Does intersect.
+			3: NewListPostings([]storage.SeriesRef{29, 30, 31}),    // Does intersect (same again).
+			4: NewListPostings([]storage.SeriesRef{60}),            // Does not intersect.
+			5: NewListPostings([]storage.SeriesRef{45}),            // Does intersect.
+			6: EmptyPostings(),                                     // Does not intersect.
+		}
+
+		indexes, err := FindIntersectingPostings(p, candidates)
+		require.NoError(t, err)
+		sort.Ints(indexes)
+		require.Equal(t, []int{1, 2, 3, 5}, indexes)
+	})
+
+	t.Run("no intersections", func(t *testing.T) {
+		p := NewListPostings([]storage.SeriesRef{10, 15, 20, 25, 30, 35, 40, 45, 50})
+		candidates := []Postings{
+			0: NewListPostings([]storage.SeriesRef{7, 13, 14, 27}), // Does not intersect.
+			1: NewListPostings([]storage.SeriesRef{60}),            // Does not intersect.
+			2: EmptyPostings(),                                     // Does not intersect.
+		}
+
+		indexes, err := FindIntersectingPostings(p, candidates)
+		require.NoError(t, err)
+		require.Empty(t, indexes)
+	})
+
+	t.Run("p is ErrPostings", func(t *testing.T) {
+		p := ErrPostings(context.Canceled)
+		candidates := []Postings{NewListPostings([]storage.SeriesRef{1})}
+		_, err := FindIntersectingPostings(p, candidates)
+		require.Error(t, err)
+	})
+
+	t.Run("one of the candidates is ErrPostings", func(t *testing.T) {
+		p := NewListPostings([]storage.SeriesRef{1})
+		candidates := []Postings{
+			NewListPostings([]storage.SeriesRef{1}),
+			ErrPostings(context.Canceled),
+		}
+		_, err := FindIntersectingPostings(p, candidates)
+		require.Error(t, err)
+	})
+
+	t.Run("one of the candidates fails on nth call", func(t *testing.T) {
+		p := NewListPostings([]storage.SeriesRef{10, 20, 30, 40, 50, 60, 70})
+		candidates := []Postings{
+			NewListPostings([]storage.SeriesRef{7, 13, 14, 27}),
+			&postingsFailingAfterNthCall{2, NewListPostings([]storage.SeriesRef{29, 30, 31, 40})},
+		}
+		_, err := FindIntersectingPostings(p, candidates)
+		require.Error(t, err)
+	})
+
+	t.Run("p fails on the nth call", func(t *testing.T) {
+		p := &postingsFailingAfterNthCall{2, NewListPostings([]storage.SeriesRef{10, 20, 30, 40, 50, 60, 70})}
+		candidates := []Postings{
+			NewListPostings([]storage.SeriesRef{7, 13, 14, 27}),
+			NewListPostings([]storage.SeriesRef{29, 30, 31, 40}),
+		}
+		_, err := FindIntersectingPostings(p, candidates)
+		require.Error(t, err)
+	})
+}
+
+type postingsFailingAfterNthCall struct {
+	ttl int
+	Postings
+}
+
+func (p *postingsFailingAfterNthCall) Seek(v storage.SeriesRef) bool {
+	p.ttl--
+	if p.ttl <= 0 {
+		return false
+	}
+	return p.Postings.Seek(v)
+}
+
+func (p *postingsFailingAfterNthCall) Next() bool {
+	p.ttl--
+	if p.ttl <= 0 {
+		return false
+	}
+	return p.Postings.Next()
+}
+
+func (p *postingsFailingAfterNthCall) Err() error {
+	if p.ttl <= 0 {
+		return errors.New("ttl exceeded")
+	}
+	return p.Postings.Err()
+}
+
+func TestPostingsWithIndexHeap(t *testing.T) {
+	t.Run("iterate", func(t *testing.T) {
+		h := postingsWithIndexHeap{
+			{index: 0, p: NewListPostings([]storage.SeriesRef{10, 20, 30})},
+			{index: 1, p: NewListPostings([]storage.SeriesRef{1, 5})},
+			{index: 2, p: NewListPostings([]storage.SeriesRef{25, 50})},
+		}
+		for _, node := range h {
+			node.p.Next()
+		}
+		heap.Init(&h)
+
+		for _, expected := range []storage.SeriesRef{1, 5, 10, 20, 25, 30, 50} {
+			require.Equal(t, expected, h.at())
+			require.NoError(t, h.next())
+		}
+		require.True(t, h.empty())
+	})
+
+	t.Run("pop", func(t *testing.T) {
+		h := postingsWithIndexHeap{
+			{index: 0, p: NewListPostings([]storage.SeriesRef{10, 20, 30})},
+			{index: 1, p: NewListPostings([]storage.SeriesRef{1, 5})},
+			{index: 2, p: NewListPostings([]storage.SeriesRef{25, 50})},
+		}
+		for _, node := range h {
+			node.p.Next()
+		}
+		heap.Init(&h)
+
+		for _, expected := range []storage.SeriesRef{1, 5, 10, 20} {
+			require.Equal(t, expected, h.at())
+			require.NoError(t, h.next())
+		}
+		require.Equal(t, storage.SeriesRef(25), h.at())
+		node := heap.Pop(&h).(postingsWithIndex)
+		require.Equal(t, 2, node.index)
+		require.Equal(t, storage.SeriesRef(25), node.p.At())
+	})
 }
